/* eslint-disable new-cap */

const Emitter = require('./Emitter');

const BaseCollection = Emitter.extend(Array);

/**
 * A class for collections of models
 * @alias Collection
 * @memberof dlx.module:base
 * @extends Array
 * @extends dlx.module:base.Emitter
 * @type {Function}
 */
class Collection extends BaseCollection {
  /**
   * Create a new Collection
   * @type {Array}
   * @param {Array}  [data = []]    An array of models to create the Collection from.
   * @param {Object} [options = {}] An options object
   * @param {Object} [options.model] The constructor for the model to use in this collection. If provided, each item in the collection will become an instance of this model.
   */
  constructor(data = [], options = {}) {

    // VALIDATION
    if (data instanceof Collection) return data;

    if (!(Array.isArray(data) || data instanceof Set)) {
      throw new Error(`The data passed to the Collection constructor must be an Array.`);
    }

    // INSTANTIATION
    const Model = options.model || Object;
    const items = Array.from(data).map(item => new Model(item));

    super(...items);

    this.model = Model;

    // VALIDATION FOR PROPERTIES AND METHODS
    const handler = {
      // NB: don't use shorthand property syntax here, in order to keep 'this' value the same
      set: (target, prop, val, proxy) => {

        if (Number.isInteger(Number(prop))) {
          return Reflect.set(target, prop, new this.model(val), proxy);
        }

        return Reflect.set(target, prop, val, proxy);

      },
    };

    return new Proxy(this, handler);

  }

  fill(val, start, end) {
    super.fill(new this.model(val), start, end);
  }

  push(...args) {
    super.push(...args.map(item => new this.model(item)));
  }

  splice(start, deleteCount, ...args) {
    super.splice(start, deleteCount, ...args.map(item => new this.model(item)));
  }

  unshift(...args) {
    super.unshift(...args.map(item => new this.model(item)));
  }

}

module.exports = Collection;
