/**
 * An Event Emitter for DOM Nodes. Accepts a single DOM Node as input, and allows events to be registered on the Node with the usual emitter methods (`.on()`, `.off()`, etc.), except for `.emit()` (since the DOM emits the event automatically). This allows for easy registering and deregistering of events. The event map is stored directly on the DOM node, so that it isn't necessary to retain a reference to the element.
 * @alias DOMListener
 * @memberof dlx.module:base
 * @type {Function}
 */
class DOMListener {
  /**
   * Create a new DOMListener
   * @param {Object|String} el The DOM Node or DOM selector string to use for this element
   */
  constructor(el) {

    if (!(el instanceof Node || typeof el === 'string')) {
      throw new Error(`The "el" attribute must be a DOM Node or a DOM selector string.`);
    }

    Object.defineProperty(this, 'el', {
      configurable: false,
      enumerable:   true,
      value:        el instanceof Node ? el : document.querySelector(el),
      writable:     false,
    });

    const events = new Map;

    if (!this.el.events) {
      Object.defineProperty(this.el, 'events', {
        configurable: false,
        enumerable:   true,
        value:        events,
        writable:     false,
      });
    }

    this.events = this.el.events;

    this.el.off  = this.off.bind(this);
    this.el.on   = this.on.bind(this);
    this.el.once = this.once.bind(this);

  }

  /**
   * Unsubscribe one or more listeners from the element.
   * @param  {String}   [event]    The event to remove. If none is specified, all events will be removed.
   * @param  {Function} [callback] The callback to remove. If none is specified, all callbacks for the specified event will be removed.
   * @param  {Object}   [options]  The options for the event listener. See the [options object of the DOM's addEventListener method]{@link https://developer.mozilla.org/en-US/docs/Web/API/EventTarget/addEventListener}.
   * @return {Object}              Returns the DOM Listener object.
   */
  off(event, callback, options) {

    if (callback) {

      const listeners = this.events.get(event);

      if (listeners) {
        this.el.removeEventListener(event, callback, options);
        listeners.delete(callback);
        if (listeners.size === 0) this.events.delete(event);
      }

    } else if (event) {

      const listeners = this.events.get(event);

      if (listeners) {
        listeners.forEach(listener => this.el.removeEventListener(event, listener, options));
      }

      this.events.delete(event);

    } else {

      this.events.forEach((listeners, ev) => {

        listeners.forEach(cb => {
          this.el.removeEventListener(ev, cb, options);
        });

      });

      this.events.clear();

    }

    return this;

  }

  /**
   * Subscribe a callback to an event
   * @param  {String}   event     The name of the event to subscribe to. Must be a valid [event type]{@link https://developer.mozilla.org/en-US/docs/Web/Events}.
   * @param  {Function} callback  The function to invoke when the event is fired
   * @param  {Object}   [options] The options for the event listener. See the [options object of the DOM's addEventListener method]{@link https://developer.mozilla.org/en-US/docs/Web/API/EventTarget/addEventListener}.
   * @return {Object}             Returns the DOM Listener object
   */
  on(event, callback, options) {
    if (!this.events.has(event)) this.events.set(event, new Set);
    this.el.addEventListener(event, callback, options);
    this.events.get(event).add(callback);
    return this;
  }

  /**
   * Subscribe a callback to an event to fire exactly once. After the callback is invoked, it will be removed from the listeners.
   * @param  {String}   event    The name of the event to subscribe to. Must be a valid [event type]{@link https://developer.mozilla.org/en-US/docs/Web/Events}.
   * @param  {Function} callback The function to invoke when the event is fired.
   * @return {Object}            Returns the DOM Listener object
   */
  once(event, callback) {

    const triggerAndRemove = () => {
      callback();
      this.off(event, callback);
    };

    this.on(event, triggerAndRemove, { once: true });

    return this;

  }

}

module.exports = DOMListener;
