class Super {}

/**
 * A class representing an Event Emitter
 * @alias Emitter
 * @memberof dlx.module:base
 * @type {Function}
 *
 * @example
 * // The Model class will inherit from Emitter
 * class Model extends Emitter {
 *   ...
 * }
 *
 * @example
 * // The Collection class will have both Array and Emitter methods, and can be subclassed further
 * class Collection = Emitter.extend(Array) {
 *   ...
 * }
 */
class Emitter {
  /**
   * Create a new Event Emitter
   * @type {Object}
   */
  constructor() {

    Object.defineProperty(this, 'events', {
      configurable: false,
      enumerable:   false,
      value:        new Map,
      writable:     false,
    });

  }

  /**
   * Triggers the callbacks for a given event. Subsequent arguments will be passed to the event callbacks.
   * @param  {String} event  The event name to trigger
   * @param  {Any}    [args] Additional arguments to pass to the event callback
   * @return {Object}        Returns a Promise that resolves when all callbacks are complete
   */
  emit(event, ...args) {
    if (typeof event !== 'string') throw new TypeError('The event name must be a string.');
    const listeners = this.events.get(event);
    if (listeners) return Promise.all(Array.from(listeners).map(cb => cb(...args)));
    return Promise.resolve();
  }

  /**
   * Unsubscribe from one or more events on this object
   * @param {String}   [event]    The event to remove. If none is specified, all events will be removed.
   * @param {Function} [callback] The callback to remove. If none is specified, all callbacks for the specified event will be removed.
   * @return {Object}             Returns the Emitter object
   */
  off(event, callback) {

    if (callback) {

      const listeners = this.events.get(event);

      if (listeners) {
        listeners.delete(callback);
        if (listeners.size === 0) this.events.delete(event);
      }

    } else if (event) {

      this.events.delete(event);

    } else {

      this.events.clear();

    }

    return this;

  }

  /**
   * Subscribe to an event on this object, by binding a callback function to the name of the event. Callbacks may be Promises or regular functions.
   * @param  {String|Object} event    The name of the event to subscribe to. May also be an object mapping event names to callback functions
   * @param  {Function}      callback The callback function to invoke whenever the specified event is fired
   * @return {Object}                 Returns the emitter object
   */
  on(event, callback) {

    const addEvent = (ev, cb) => {

      if (typeof cb !== 'function') {
        throw new TypeError('Callback must be a function.');
      }

      if (!this.events.has(ev)) this.events.set(ev, new Set);
      this.events.get(ev).add(cb);

    };

    if (typeof event === 'object') {

      for (const ev in event) {
        if (event.hasOwnProperty(ev)) {
          addEvent(ev, event[ev]);
        }
      }

    } else {

      addEvent(event, callback);

    }

    return this;

  }

  /**
   * Subscribe to an event on this object for one time only. After the event is triggered, the callback will be removed. Callbacks may be Promises or regular functions.
   * @param  {String}   event    The name of the event to subscribe to
   * @param  {Function} callback The callback function to invoke when the event is fired
   * @return {Object}            Returns the emitter object
   */
  once(event, callback) {

    if (typeof event !== 'string') {
      throw new TypeError('The event argument must be a string.');
    }

    if (typeof callback !== 'function') {
      throw new TypeError('The callback must be a function.');
    }

    const triggerAndRemove = () => {
      callback();
      const listeners = this.events.get(event);
      listeners.delete(triggerAndRemove);
      if (listeners.size === 0) this.events.delete(event);
    };

    this.on(event, triggerAndRemove);

    return this;

  }

  /**
   * Mixin the Event Emitter into another class. This method takes the provided class, adds the Emitter methods to it, and returns a new class that you can then subclass further.
   * @param {Function} [Superclass] A class constructor to mixin the Emitter with
   * @return {Function}             Returns the new class with the Emitter methods mixed in
   * @example
   * const Collection = Emitter.extend(Array);
   */
  static extend(Superclass = Super) {

    return class extends Superclass {
      constructor(...args) {
        super(...args);
        const props = Object.getOwnPropertyDescriptors(Emitter.prototype);
        Object.defineProperties(this, props);
      }
    };

  }

}

module.exports = Emitter;
