const Emitter = require('./Emitter');

/**
 * A base Model that other models inherit from. It also has a few static methods that are useful in creating subclasses.
 * @alias Model
 * @memberof dlx.module:base
 * @type {Object}
 * @extends dlx.module:base.Emitter
 */
class Model extends Emitter {

  /**
   * A no-op destroy function. This method should be overwritten on the subclass with logic to delete the model from the database.
   */
  destroy() {}

  /**
   * A no-op save function. This method should be overwritten on the subclass with logic to save the model data to a database.
   */
  save() {}

  /**
   * Returns a JSON representation of this model, as a string
   * @return {String} The JSON string
   */
  get json() {
    return JSON.stringify(this.toJSON(), null, 2);
  }

  /**
   * Adds an array property to the provided object with the given property name. The array property will be enumerable but non-configurable and non-writable. Each item added to the array will be made an instance of the provided model.
   * @param {Object}   object              The object to add the array property to
   * @param {String}   propName            The name of the array property to add
   * @param {Function} model               A constructor function to apply to each item before being added to the array.
   * @param {Array}    [initialValue = []] The intial set of items to add to the array property.
   */
  static defineArray(object, propName, model, initialValue = []) {

    if (!Array.isArray(initialValue)) {
      throw new TypeError(`The "${propName}" property must be an array.`);
    }

    const ItemModel = model;
    const isString  = ItemModel === String;

    const handler = {
      set(target, prop, val, proxy) {

        let newVal = val;

        if (Number.isInteger(Number(prop))) {
          if (isString && typeof val !== 'string') {
            throw new TypeError(`Each of the items in "${propName}" must be a String.`);
          } else if (!isString) {
            newVal = new ItemModel(val);
          }
        }

        return Reflect.set(target, prop, newVal, proxy);

      },
    };

    const p = new Proxy([], handler);

    p.push(...initialValue);

    Object.defineProperty(object, propName, {
      configurable: true,
      enumerable:   true,
      value:        p,
      writable:     false,
    });

  }

  /**
   * Adds a property to the provided object which may only have the specified values
   * @param  {Object} object       The object to add the property to
   * @param  {String} propName     The name of the property to add
   * @param  {Array}  values       An array of acceptable values for the new property
   * @param  {Any}    initialValue The initial value to give to the new property. Will be `undefined` if none is provided.
   * @return {Object}              Returns the original object with the new property added
   */
  static defineEnumProp(object, propName, values, initialValue) {

    if (!(object instanceof Object)) {
      throw new TypeError(`The "object" argument must be an Object.`);
    }

    if (!Array.isArray(values)) {
      throw new TypeError(`The "values" argument must be an array of allowed values.`);
    }

    const o = object;

    let value;

    Object.defineProperty(o, propName, {
      configurable: true,
      enumerable:   true,
      get() { return value; },
      set(val) {

        if (!values.includes(val)) {
          throw new Error(`The value provided is not a valid value for the "${propName}" property.`);
        }

        value = val;

      },
    });

    if (typeof initialValue !== 'undefined') o[propName] = initialValue;

    return o;

  }

  /**
   * Adds a property with the given name to the provided object, and ensures that the value of the property is always an instance of the provided model. The property will be enumerable and writable but non-configurable.
   * @param {Object}   object         The object to add the property to
   * @param {String}   propName       The name of the property to add
   * @param {Function} model          A constructor function. The value of this property will always be an instance of the provided function.
   * @param {Any}      [initialValue] The initial value to use for the value of this property. If none is provided, the value will be undefined until the property is set.
   * @return {Object}                 Returns the original object with the new property added
   */
  static defineProp(object, propName, model, initialValue) {

    if (!(object instanceof Object)) throw new TypeError(`The "object" argument must be an Object.`);

    const o         = object;
    const PropModel = model;
    const isDate    = PropModel === Date;
    const isString  = PropModel === String;

    let value;

    Object.defineProperty(o, propName, {
      configurable: true,
      enumerable:   true,
      get() { return value; },
      set(val) {

        if (isString) {

          if (typeof val !== 'string') {
            throw new TypeError(`The "${propName}" property must be a String.`);
          }

          value = val;

        } else if (isDate) {

          if (isNaN(Date.parse(val))) {
            throw new TypeError(`The "${propName}" property must be a valid Date string or object.`);
          }

          value = new Date(val);

        } else if (typeof val !== 'undefined') {

          value = new PropModel(val);

        }

      },
    });

    if (typeof initialValue !== 'undefined') o[propName] = initialValue;

    return o;

  }

  /**
   * Takes an Object and an Array of required properties, and returns a simplified version of the object with all empty Arrays, Strings, Sets, Maps, and basic Objects removed. This is useful when trying to reduce the footprint of the object when stringifying it.
   * @param  {Object} model           The Object to simplify
   * @param  {Array}  [required = []] A Array of required property names. These properties will not be removed from the object, even if they are empty
   * @return {Object}                 Returns a *new* object with the properties removed
   */
  static simplify(model, required = []) {

    if (!(model instanceof Object)) {
      throw new TypeError(`The "model" argument must be an Object.`);
    }

    if (!Array.isArray(required) || required.some(val => typeof val !== 'string')) {
      throw new TypeError(`The "required" argument must be an Array of Strings.`);
    }

    const o = Object.assign({}, model);

    for (const prop in model) {

      if (!required.includes(prop)) {

        if (
          (Array.isArray(o[prop]) && !o[prop].length)
          || ((o[prop] instanceof Set || o[prop] instanceof Map) && !o[prop].size)
          || (o[prop] && o[prop].constructor === Object && !Object.keys(o[prop]).length)
          || (typeof o[prop] !== 'number' && !o[prop])
        ) {
          delete o[prop];
        }

      }

    }

    return o;

  }

}

module.exports = Model;
