const { toDOM }   = require('../utilities');
const Emitter     = require('./Emitter');
const DOMListener = require('./DOMListener');

/**
 * A base View class that other views inherit from
 * @alias View
 * @memberof dlx.module:base
 * @extends  dlx.module:base.Emitter
 * @type {Object}
 * @prop {Object}   el        The DOM element attached to this view (this Node exists whether or not it has been inserted into the DOM). This may be either the container element where you'll insert HTML, or the actual element to insert into the DOM. Will be an empty div if no element was provided during intialization.
 * @prop {Object}   model     The data for this view
 * @prop {String}   position  The position within the container element to insert this view
 * @prop {Function} template  The templating function for this view. May return either an HTML string or a single DOM Node.
 */
class View extends Emitter {
  /**
   * Create a new View
   * @type  {Function}
   * @param {Object}        [data = {}]                The model or raw data to create the view from
   * @param {Object}        [options = {}]             An options object.
   * @param {Object|String} [options.container]        The container to insert the view into. May be either a DOM Node or a DOM selector string. If provided, the `.render()` method may be called without the `container` option.
   * @param {Object|String} [options.el]               If provided, this option sets the value of the `el` property of the View. May be either a DOM Node or a DOM selector string. This is typically either the container element that the render method will insert HTML in, or the actual element to be inserted. If this option is provided, the `template` method will not be invoked.
   * @param {String}        [options.position = 'end'] The position to render this view at within its container. Possible value are `start` (inserts the view as the first element in the container), `end` (inserts the view as the last element in the container), and `replace` (clears all child nodes from the container before inserting this view). If provided, the `.render()` method may be called without the `position` option provided.
   * @param {Boolean}       [options.template]         A templating function that accepts the view's model as an input, and returns either an HTML string or DOM Node. This method may also simply be provided as a method on the view subclass.
   */
  constructor(data = {}, options = {}) {

    super();

    this.model    = data;
    this.options  = options;
    this.position = options.position || 'end';
    this.template = options.template || this.template;

    if (options.container) {
      this.container = options.container instanceof Node ?
        options.container :
        document.querySelector(options.container);
    }

    let el;

    Object.defineProperty(this, 'el', {
      configurable: false,
      enumerable:   true,
      get() { return el; },
      set(val) {
        el = new DOMListener(val).el;
        return el;
      },
    });

    if (options.el) {
      this.el = options.el instanceof Node ? options.el : document.querySelector(options.el);
    }

    this.el = this.el || toDOM(this.template(data))[0] || document.createElement('div');

  }

  /**
   * Clears the container element of any HTML and removes all listeners from this View, without removing the container element from the DOM
   * @return {Object} Returns the view object
   */
  clear() {
    this.container.innerHTML = '';
    this.off();
    return this;
  }

  /**
   * Removes all listeners from this View, and removes its container element from the DOM
   * @return {Object} Returns the view object
   */
  remove() {
    this.el.remove();
    this.off();
    return this;
  }

  /**
   * Inserts this view into the DOM, within the specified container. You will often want to overwrite this function on the subclass.
   * @param {Object}        [options = {}]             An options object
   * @param {Object|String} [options.container]        A container element to render the view in. May be either a DOM selector string or a DOM Node. If this option was already provided in the view's `container` option, this argument may be omitted.
   * @param {String}        [options.position = 'end'] The position to render this view at within its container. Possible value are `start` (inserts the view as the first element in the container), `end` (inserts the view as the last element in the container), and `replace` (clears all child nodes from the container before inserting this view). If this option was already provided in the view's `position` option, this argument may be omitted.
   * @return {Object}                          Returns this view object
   */
  render(options = {}) {

    if (typeof options !== 'object') throw new Error(`Render options must be an object.`);

    const container = options.container || this.container;
    const position  = options.position || this.position;

    if (!(options.container instanceof Node || typeof options.container === 'string')) {
      throw new TypeError(`The "container" option must be a DOM Node or a DOM selector string.`);
    }

    const wrapper   = container instanceof Node ? container : document.querySelector(container);

    switch (position) {

      case 'start': {
        const { firstChild } = wrapper;
        wrapper.insertBefore(this.el, firstChild);
        break;
      }

      case 'replace': {
        wrapper.innerHTML = '';
        wrapper.appendChild(this.el);
        break;
      }

      default: wrapper.appendChild(this.el);

    }

    return this;

  }

  /**
   * A no-op template function, to be replaced by a template function on the view subclass. The template function should accept the model data as the first argument, and return either an HTML string or a DOM Node. This may also be provided in the `template` option.
   */
  template() {}

}

module.exports = View;
