const { Model }       = require('../base');
const MultiLangString = require('./MultiLangString');

const defaults = {
  environments:  [],
  transcription: {},
};

/**
 * A class representing an Allomorph of a Lexeme
 * @alias Allomorph
 * @memberof dlx.module:models
 * @prop {Set}             environments  A set of the (morpho)phonological environments for this allomorph
 * @prop {MultiLangString} transcription The transcription of this allomorph, as a [MultiLangString]{@link http://developer.digitallinguistics.io/spec/schemas/MultiLangString.html}
 * @prop {String}          txn           A shorthand for accessing the default orthography of the `transcription` object, if one is set
 * @example
 * const allomorph = new Allomorph({
 *   environments:  ['_i', '_y'],
 *   transcription: { ipa: 't͡ʃ' }
 * });
 *
 * console.log(allomorph.transcription.ipa);      // t͡ʃ
 * console.log(allomorph.environments.has('_i')); // true
 */
class Allomorph {
  /**
   * Create a new Allomorph
   * @param {Object}          [data]             The data to use for the Allomorph
   * @param {Array|Set}       data.environments  An Array or Set of environments, as strings. May be empty.
   * @param {MultiLangString} data.transcription A transcription of this allomorph, optionally in multiple orthographies. Do not include any leading or trailing tokens (e.g. hyphens, equal signs). Should be formatted as a [MultiLangString]{@link http://developer.digitallinguistics.io/spec/schemas/MultiLangString.html}.
   */
  constructor(data = defaults) {

    // VALIDATION
    if (data instanceof Allomorph) return data;

    if (typeof data !== 'object') {
      throw new Error(`The data passed to the Allomorph constructor must be an object.`);
    }

    if (data.environments && !(Array.isArray(data.environments) || data.environments instanceof Set)) {
      throw new Error(`The "environements" property must be an Array or Set.`);
    }

    // INSTANTIATION
    Object.assign(this, data);

    Model.defineProp(this, 'transcription', MultiLangString, data.transcription || {});

    MultiLangString.alias(this, 'transcription', 'txn');

    const environments = new Set;

    // proxies environments.add for validation when new items are added
    Object.defineProperty(environments, 'add', {
      configurable: false,
      enumerable:   false,
      value:        new Proxy(environments.add, {
        apply(target, context, [val]) {
          if (typeof val !== 'string') {
            throw new TypeError(`The value to add to "environments" must be a String.`);
          }
          return Reflect.apply(target, context, [val]);
        },
      }),
      writable:     false,
    });

    // prevents 'environments' from being overwritten, which would lose its proxied `add` method
    Object.defineProperty(this, 'environments', {
      configurable: false,
      enumerable:   true,
      value:        environments,
      writable:     false,
    });

    Array.from(data.environments || []).forEach(env => this.environments.add(env));

  }

  toJSON() {
    const data = Model.simplify(this, [
      'environments',
      'transcription',
    ]);
    data.environments = Array.from(this.environments);
    return data;
  }

}

module.exports = Allomorph;
