const { checkProp }   = require('../utilities');
const { Model }       = require('../base');
const Location        = require('./Location');
const MultiLangString = require('./MultiLangString');
const Orthography     = require('./Orthography');
const Phoneme         = require('./Phoneme');

const checkGlottolog = val => {
  if (!/[a-z]{4}[0-9]{4}/.test(val)) {
    throw new Error(`The Glottolog code must be formatted as a String of 4 letters followed by 4 numbers.`);
  }
};

const checkISO = val => {
  if (!/^[a-z]{3}$/.test(val)) {
    throw new Error(`The ISO code must be formatted as a 3-letter String.`);
  }
};

const checkNames = names => {

  if (!(Array.isArray(names) || names instanceof Set)) {
    throw new TypeError(`The "additionalNames" property must be an Array or Set.`);
  }

  names.forEach(val => {
    if (typeof val !== 'string') {
      throw new TypeError(`Each item in the "additionalNames" property must be a String.`);
    }
  });

};

/**
 * A class representing a Language
 * @alias Language
 * @memberof dlx.module:models
 * @extends Model
 * @prop {String}          abbreviation    An abbreviation for this Language, as a String.
 * @prop {Set}             additionalNames A Set of additional names for this language, as Strings.
 * @prop {Date}            dateCreated     The date and time that this Language object was created
 * @prop {Date}            dateModified    The date and time that this Language object was last modified. Setting any property on this object will update this property automatically.
 * @prop {Array}           locations       An Array of locations where this language variety is spoken, as [Location objects]{@link dlx.module:models.Location}.
 * @prop {String}          glottolog       The Glottolog code for this language variety
 * @prop {String}          iso             The ISO 639-3 code for this language variety
 * @prop {MultiLangString} name            The name of this language, as a [MultiLangString]{@link dlx.module:models.MultiLangString}.
 * @prop {Array}           orthographies   An Array of orthographies used by this language, each as an [Orthography object]{@link dlx.module:models.Orthography}
 * @prop {Array}           phonemes        An Array of phonemes in this language, as [Phoneme objects]{@link dlx.module:models.Phoneme}.
 * @prop {URL}             url             The URL where the JSON data for this language may be retrieved, as a valid URI
 */
class Language extends Model {
  /**
   * Create a new Language
   * @param {Object}      [data]                 The data to use for this Language object
   * @param {String}      [data.abbreviation]    An abbreviation for this language, formatted as a valid [Abbreviation]{@link http://developer.digitallinguistics.io/spec/schemas/Abbreviation.html}.
   * @param {Array|Set}   [data.additionalNames] An Array or Set of additional names for this language, as Strings.
   * @param {Object}      [data.codes]           An Object containing two properties, `glottolog` and `iso`, each of which must be Arrays or Sets of Strings.
   * @param {Date|String} [data.dateCreated]     The date and optionally time that the data for this Language object was originally created, as either a Date object or a valid date string. If none is provided, this property will be set automatically upon instantiation.
   * @param {Date|String} [data.dateModified]    The date and optionally time that the data for this Language object was last modified, as either a Date object or a valid date string. If none is provided, this property will be set automatically upon instantiation.
   * @param {String}      [data.glottolog]       The Glottolog code for this language variety
   * @param {String}      [data.iso]             The ISO 639-3 code for this language variety
   * @param {Array}       [data.locations]       An Array of locations where this language variety is spoken, each formatted as a [Location object]{@link http://developer.digitallinguistics.io/spec/schemas/Location.html}.
   * @param {Object}      data.name              The name of this language, formatted as a [MultiLangString]{@link http://developer.digitallinguistics.io/spec/schemas/MultiLangString.html}.
   * @param {Array}       [data.orthographies]   An Array of orthographies used by this language, each formatted as an [Orthography object]{@link http://developer.digitallinguistics.io/spec/schemas/Orthography.html}
   * @param {Array}       [data.phonemes]        An Array of phonemes in this language, formatted as [Phoneme objects]{@link http://developer.digitallinguistics.io/spec/schemas/Phoneme.html}.
   * @param {URL}         [data.url]             The URL where the JSON data for this language may be retrieved, as a valid URI
   */
  constructor(data = {}) {

    // VALIDATION
    if (!(data instanceof Object)) {
      throw new TypeError(`The data passed to the Language constructor must be an Object.`);
    }

    if ('abbreviation' in data) checkProp(data.abbreviation, 'abbr', 'abbreviation');

    if ('additionalNames' in data) checkNames(data.additionalNames);

    if ('glottolog' in data) checkGlottolog(data.glottolog);

    if ('iso' in data) checkISO(data.iso);

    if ('url' in data) checkProp(data.url, 'url', 'url');

    // INITIALIZATION
    super();

    Object.assign(this, data);

    this.additionalNames = new Set(data.additionalNames);

    Model.defineProp(this, 'dateCreated', Date, data.dateCreated || new Date);
    Model.defineProp(this, 'dateModified', Date, data.dateModified || new Date);
    Model.defineProp(this, 'name', MultiLangString, data.name || {});

    Model.defineArray(this, 'locations', Location, data.locations);
    Model.defineArray(this, 'orthographies', Orthography, data.orthographies);
    Model.defineArray(this, 'phonemes', Phoneme, data.phonemes);

    // PROXY VALIDATION FOR SETTERS
    const handler = {
      set(target, prop, val, proxy) {

        switch (prop) {
          case 'abbreviation': checkProp(val, 'abbr', 'abbreviation'); break;
          case 'additionalNames': checkNames(val); break;
          case 'glottolog': checkGlottolog(val); break;
          case 'iso': checkISO(val); break;
          case 'url': checkProp(val, 'url', 'url'); break;
          default: break;
        }

        target.dateModified = new Date;

        return Reflect.set(target, prop, val, proxy);

      },
    };

    return new Proxy(this, handler);

  }

  toJSON() {
    const data = Model.simplify(this, ['name']);
    if (data.additionalNames) data.additionalNames = Array.from(data.additionalNames);
    return data;
  }

}

module.exports = Language;
