const { checkProp }   = require('../utilities');
const { Model }       = require('../base');
const MultiLangString = require('./MultiLangString');

/**
 * A class representing a Lexeme Reference
 * @alias LexemeReference
 * @memberof dlx.module:models
 * @prop {String}          lexeme      The key (abbreviation) of the lexeme being referenced
 * @prop {String}          lexemeUrl   The URL of the referenced lexeme. Must be a valid URI.
 * @prop {String}          lexicon     The key (abbreviation) of the lexicon being referenced
 * @prop {String}          lexiconUrl  The URL of the referenced lexicon. Must be a valid URI.
 * @prop {String}          relation    The type of relation that holds between this item and the referenced lexeme
 * @prop {Integer}         sense       The index of the sense being referenced
 * @prop {MultiLangString} variantType The type of variant that the referenced lexeme is. A [MultiLangString]{@link dlx.module:models.MultiLangString}.
 *
 * @example
 * const ref = new LexemeReference('calle1');
 *
 * console.log(ref.lexeme); // calle1
 *
 * @example
 * const ref = new LexemeReference({
 *   lexeme:  'calle1',
 *   lexicon: 'spanish',
 * });
 *
 * console.log(ref.lexeme); // calle1
 */
class LexemeReference {
  /**
   * Create a new Lexeme Reference.
   * @param {Object|String} [data]             The data to use for the Lexeme Reference. If the data is a string, that string will become the "lexeme" property, indicating the key of the referenced lexeme. Otherwise an object must be passed containing a "lexeme" property.
   * @param {String}        data.lexeme        The key (abbreviation) of the lexeme being referenced
   * @param {String}        [data.lexemeUrl]   The URL of the referenced lexeme. Must be a valid URI.
   * @param {String}        [data.lexicon]     The key (abbreviation) of the lexicon being referenced
   * @param {String}        [data.relation]    A string indicating the type of relationship between the current item and the referenced lexeme. This property is required when the reference is being used in the "lexicalRelations" property of the Lexeme.
   * @param {Integer}       [data.sense]       The number of the sense being reference (i.e. its index in the senses array). Note that the senses array is zero-indexed.
   * @param {Object}        [data.variantType] The type of variant that the referenced lexeme is. Must be formatted as a [MultiLangString]{@link http://developer.digitallinguistics.io/spec/schemas/MultiLangString.html}.
   */
  constructor(data = '') {

    // VALIDATION
    if (data && typeof data !== 'object' && typeof data !== 'string') {

      throw new Error(`The data passed to the LexemeReference constructor must be an object or string.`);

    } else if (typeof data === 'object') {

      if ('lexemeUrl' in data) checkProp(data.lexemeUrl, 'url', 'lexemeUrl');
      if ('lexiconUrl' in data) checkProp(data.lexiconUrl, 'url', 'lexiconUrl');
      if ('relation' in data) checkProp(data.relation, 'abbr', 'relation');
      if ('sense' in data) checkProp(data.sense, 'integer', 'sense');

    }

    // INSTANTIATION
    if (typeof data === 'string') {
      this.lexeme = data;
    } else if (typeof data === 'object') {
      this.lexeme = data.lexeme || '';
      if ('lexemeUrl' in data) this.lexemeUrl     = data.lexemeUrl;
      if ('lexicon' in data) this.lexicon         = data.lexicon;
      if ('lexiconUrl' in data) this.lexiconUrl   = data.lexiconUrl;
      if ('sense' in data) this.sense             = data.sense;
      if ('relation' in data) this.relation       = data.relation;
      Model.defineProp(this, 'variantType', MultiLangString, data.variantType);
    }

    MultiLangString.alias(this, 'variantType', 'varType');

    // VALIDATION FOR SETTERS
    const handler = {
      set(target, prop, val, proxy) {

        switch (prop) {
          case 'lexeme':
          case 'lexicon':
          case 'relation':
            checkProp(val, 'abbr', prop);
            break;
          case 'lexemeUrl':
          case 'lexiconUrl':
            checkProp(val, 'url', prop);
            break;
          case 'sense': checkProp(val, 'integer', 'sense'); break;
          default: return false;
        }

        return Reflect.set(target, prop, val, proxy);

      },
    };

    return new Proxy(this, handler);

  }

  toJSON() {
    return Model.simplify(this, ['lexeme']);
  }

}

module.exports = LexemeReference;
