const { Model }       = require('../base');
const LexemeReference = require('./LexemeReference');
const MultiLangString = require('./MultiLangString');
const Note            = require('./Note');
const Tags            = require('./Tags');

const defaults = {
  gloss: {},
  transcription: {},
};

/**
 * A class representing a Morpheme Token
 * @alias Morpheme
 * @memberof dlx.module:models
 * @prop {MultiLangString} gloss         The gloss for this morpheme, as a [MultiLangString]{@link dlx.module:models:MultiLangString}.
 * @prop {String}          gl            A shorthand for accessing the default orthography of the `gloss` property.
 * @prop {Object}          lexeme        A reference to an item in a lexicon where more information about this morpheme can be found, as a [LexemeReference]{@link dlx.module:models.LexemeReference}.
 * @prop {Array}           notes         An array of notes about this morpheme. Each item should be a [Note object]{@link dlx.module:models.Note}.
 * @prop {Tags}            tags          A set of tags for this morpheme, as a [Tags object]{@link dlx.module:models.Tags}.
 * @prop {MultiLangString} transcription A transcription of this morpheme, as a [MultiLangString]{@link dlx.module:models.MultiLangString}.
 * @prop {String}          txn           A shorthand for accessing the default orthography of the `transcription` property.
 */
class Morpheme extends Model {
  /**
   * Create a new Morpheme
   * @param {Object} [data = {}]          The data to use for this Morpheme Token
   * @param {Object} data.gloss           The gloss for this morpheme, formatted as a [MultiLangString]{@link http://developer.digitallinguistics.io/spec/schemas/MultiLangString.html}.
   * @param {Object} [data.lexeme]        A Lexeme Reference that points to an item in a lexicon where more information about this morpheme can be found.
   * @param {Array}  [data.notes]         An array of notes about this Morpheme. Each note should be formatted as a [Note object]{@link http://developer.digitallinguistics.io/spec/schemas/Note.html}.
   * @param {Object} [data.tags]          A set of tags for this morpheme, formatted as a [Tags object]{@link http://developer.digitallinguistics.io/spec/schemas/Tags.html}.
   * @param {Object} data.transcription   The transcription of this morpheme, formatted as a [MultiLangString]{@link http://developer.digitallinguistics.io/spec/schemas/MultiLangString.html}.
   */
  constructor(data = defaults) {

    // VALIDATION
    if (!(data instanceof Object)) {
      throw new TypeError(`The "data" argument must be an Object.`);
    }

    // INSTANTIATION
    super();

    Object.assign(this, data);

    Model.defineProp(this, 'gloss', MultiLangString, data.gloss || {});
    Model.defineProp(this, 'lexeme', LexemeReference, data.lexeme || {});
    Model.defineProp(this, 'tags', Tags, data.tags || {});
    Model.defineProp(this, 'transcription', MultiLangString, data.transcription || {});

    Model.defineArray(this, 'notes', Note, data.notes);

    MultiLangString.alias(this, 'gloss', 'gl');
    MultiLangString.alias(this, 'transcription', 'txn');

  }

  toJSON() {
    return Model.simplify(this, [
      'gloss',
      'transcription',
    ]);
  }

}

module.exports = Morpheme;
