const { isAbbr } = require('../utilities');
const { Model }  = require('../base');

// validates a writing system key
const validateKey = key => {
  if (!isAbbr(key)) {
    throw new Error(`Each writing system key must be a valid Abbreviation.`);
  }
};

// validates a transcription string
const validateString = string => {
  if (typeof string !== 'string') {
    throw new Error(`Each transcription must be a string.`);
  }
};

/**
 * A class representing a transcription in one or more languages or writing systems.
 * @alias MultiLangString
 * @memberof dlx.module:models
 * @extends Map
 * @prop {String} default The default orthography to get and set. This property is not enumerable.
 * @example
 * const str = new MultiLangString({
 *   spa: 'Hola, me llamo Daniel.',
 *   ipa: 'ola me jamo danjɛl',
 * }, {
 *   default: 'spa',
 * });
 *
 * str.eng = 'Hello, my name is Daniel.';
 *
 * console.log(str.spa);     // Hola, me llamo Daniel.
 * console.log(str.ipa);     // ola me jamo danjɛl
 * console.log(str.default); // 'spa'
 */
class MultiLangString extends Map {
  /**
   * Create a new Multi-Language String
   * @param {Object}  [data]            The data for this transcription, as a plan Object or Map. See the [MultiLangString specification]{@link http://developer.digitallinguistics.io/spec/schemas/MultiLangString.html} for more information on formatting transcription data.
   * @param {Object}  [options]         An options object
   * @param {String}  [options.default] The default orthography to get and set
   */
  constructor(data = {}, options = {}) {

    // VALIDATION
    if (data instanceof Map) return data;

    if (typeof data !== 'object') {
      throw new TypeError(`The data passed to the MultiLangString constructor must be an object.`);
    }

    if (typeof options !== 'object') {
      throw new TypeError(`The options argument must be an object.`);
    }

    Object.keys(data).forEach(validateKey);
    Object.values(data).forEach(validateString);

    // INSTANTIATION
    super(Object.entries(data));

    Model.defineProp(this, 'default', String, options.default);

    // VALIDATION FOR SETTERS
    const handler = {
      get(target, prop) {
        if (target[prop] instanceof Function) return target[prop].bind(target);
        return target.get(prop) || target[prop];
      },
      set(target, prop, val) {
        validateKey(prop);
        validateString(val);
        return target.set(prop, val);
      },
    };

    return new Proxy(this, handler);

  }

  toJSON() {
    const data = {};
    Array.from(this.entries()).forEach(([key, val]) => { data[key] = val; });
    return data;
  }

  /**
   * Adds a new property to the provided object. The new property is an alias for the default orthography of the given MultiLangString property (i.e., a shorthand for `object[mlsProp][object[mlsProp].default]`). The new property will not be configurable or enumerable, only writable.
   * @param  {Object} object      The object to add the alias property to
   * @param  {String} mlsProp     The name of the MultiLangString property whose default orthography you want to alias
   * @param  {String} defaultProp The name of the new alias property.
   * @return {Object}             Returns the object with the new property added
   *
   * @example
   * const phrase = {
   *   translation: new MultiLangString({
   *     default: 'eng',
   *     eng:     'Hello world!',
   *     spa:     'Hola mundo!',
   *   }),
   * };
   *
   * MultiLangString.alias(phrase, 'translation', 'tln');
   *
   * console.log(phrase.tln); // 'Hello world!'
   */
  static alias(object, mlsProp, defaultProp) {

    const o = object;

    Object.defineProperty(o, defaultProp, {
      configurable: false,
      enumerable:   false,
      get() { return o[mlsProp][o[mlsProp].default]; },
      set(val) { o[mlsProp][o[mlsProp].default] = val; },
    });

    return o;

  }

}

module.exports = MultiLangString;
