const { checkProp }   = require('../utilities');
const { Model }       = require('../base');

// don't use destructuring of index.js here - causes problematic circular references
const MultiLangString = require('./MultiLangString');
const Note            = require('./Note');
const Tags            = require('./Tags');
const Word            = require('./Word');

// checks whether the phrase key is properly formatted
const checkKey = key => {
  if (!/^[(a-z)|(A-Z)|(0-9)]+\.[0-9]{1,3}$/.test(key)) {
    throw new Error(`Improperly formatted key.`);
  }
};

// check whether the end time comes before the start time
const checkTimes = (start, end) => {
  if (start > end) throw new Error(`The value of the "endTime" property must be greater than the value of the "startTime" property.`);
};

const defaultPhrase = {
  transcription: {},
  translation:   {},
  words:         [],
};

/**
 * A class representing a Phrase
 * @alias Phrase
 * @memberof dlx.module:models
 * @prop {Number}          endTime       The time within the accompanying media files that this phrase ends, in seconds (decimals allowed).
 * @prop {String}          key           A key that uniquely identifies this phrase within the text, as a string.
 * @prop {String}          language      The abbreviation for the language of this phrase.
 * @prop {Array}           notes         An array of notes about this phrase. Each note is an instance of the [Note class]{@link dlx.module:models.Note}.
 * @prop {String}          speaker       The abbreviation of the speaker of this phrase.
 * @prop {Number}          startTime     The time within the accompanying media files that this phrase begins, in seconds (decimals allowed).
 * @prop {Tags}            tags          A set of tags for this phrase, as a [Map object]{@link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Map}.
 * @prop {MultiLangString} transcription The transcriptions for this phrase, as a [MultiLangString]{@link dlx.module:models.MultiLangString}
 * @prop {String}          txn           A shorthand for accessing the default orthography of the `transcription` object, if one is set
 * @prop {MultiLangString} translation   The translations for this phrase, as a [MultiLangString]{@link dlx.module:models.MultiLangString}
 * @prop {String}          tln           A shorthand for accessing the default orthography of the `translation` object, if one is set
 * @prop {String}          url           The URL where this phrase can be accessed.
 * @prop {Array}           words         An array of word tokens in this phrase. Each item is an instance of the [Word class]{@link dlx.module:models.Word}.
 *
 * @example
 * const phrase = new Phrase({
 *
 *   transcription: {
 *     spa:  'Hola, me llamo Daniel.',
 *     ipa:  'ola me jamo dænjɛl',
 *   },
 *
 *   translation: {
 *     eng:  'Hello, my name is Daniel.',
 *     type: 'free',
 *   },
 *
 *   words: [],
 *
 * });
 */
class Phrase extends Model {
  /**
   * Create a new Phrase
   * @param {Object}     [data]                  The raw data for this phrase. See the [DLx specification for Phrases]{@link http://developer.digitallinguistics.io/spec/schemas/Phrase.html} for more information on formatting phrase data.
   * @param {Number}     [data.endTime]          The time within the accompanying media files that this phrase ends, in seconds (decimals allowed).
   * @param {String}     [data.key]              A key that uniquely identifies this phrase within the text, as a string. Must consist of a valid abbreviation, a period, and the number of this phrase within the text (index starts at 1). For example: `A.3` or `DogStory.24`.
   * @param {String}     [data.language]         The abbreviation for the language that this phrase is in. Must be a valid [Abbreviation]{@link http://developer.digitallinguistics.io/spec/schemas/abbreviation.html}.
   * @param {Array}      [data.notes]            An array of notes about this phrase. Each note should adhere to the [Note format]{@link http://developer.digitallinguistics.io/spec/schemas/Note.html}.
   * @param {String}     [data.speaker]          The abbreviation of the speaker of this phrase. Must be a valid [Abbreviation]{@link http://developer.digitallinguistics.io/spec/schemas/abbreviation.html}.
   * @param {Number}     [data.startTime]        The time within the accompanying media files that this phrase begins, in seconds (decimals allowed).
   * @param {Map|Object} [data.tags]             An object containing tags and their values. May also be an iterable object, whose items are arrays of tags and their values (in other words, the standard method for instantiating a new Map object; see [MDN's Map documentation]{@link https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Map} for more details).
   * @param {Object}     data.transcription      The transcriptions of this phrase, optionally in multiple orthographies. Must be formatted as a [MultiLangString]{@link http://developer.digitallinguistics.io/spec/schemas/MultiLangString.html}.
   * @param {Object}     data.translation        The translations of this phrase, optionally in multiple orthographies. Must be formatted as a [MultiLangString]{@link http://developer.digitallinguistics.io/spec/schemas/MultiLangString.html}. May also include a `type` property.
   * @param {String}     [data.translation.type] The type of translation (usually `free` or `literal`).
   * @param {String}     [data.url]              The URL where this phrase can be accessed. Must be a valid URI.
   * @param {Array}      data.words              An array of word tokens in this phrase. Each item must be formatted as a [Word object]{@link http://developer.digitallinguistics.io/spec/schemas/MultiLangString.html}.
   */
  constructor(data = defaultPhrase) {

    // VALIDATION
    if (data instanceof Phrase) return data;

    if (typeof data !== 'object') {
      throw new Error(`The data passed to the Phrase constructor must be an object.`);
    }

    if (!(
      ('endTime' in data && 'startTime' in data) || !('endTime' in data || 'startTime' in data)
    )) {
      throw new Error(`The "startTime" and "endTime" properties must either both be absent, or both be present.`);
    }

    if ('endTime' in data) checkProp(data.endTime, 'number', 'endTime');

    if ('key' in data) checkKey(data.key);

    if ('language' in data) checkProp(data.language, 'abbr', 'language');

    if ('speaker' in data) checkProp(data.speaker, 'abbr', 'speaker');

    if ('startTime' in data) checkProp(data.startTime, 'number', 'startTime');

    if ('url' in data) checkProp(data.url, 'url', 'url');

    if (data.startTime) checkTimes(data.startTime, data.endTime);

    // INSTANTIATION
    super();

    Object.assign(this, data);

    Model.defineProp(this, 'tags', Tags, data.tags || {});
    Model.defineProp(this, 'transcription', MultiLangString, data.transcription || {});
    Model.defineProp(this, 'translation', MultiLangString, data.translation || {});

    Model.defineArray(this, 'notes', Note, data.notes);
    Model.defineArray(this, 'words', Word, data.words);

    MultiLangString.alias(this, 'transcription', 'txn');
    MultiLangString.alias(this, 'translation', 'tln');

    if ('endTime' in data) this.endTime     = Number(data.endTime.toFixed(3));
    if ('startTime' in data) this.startTime = Number(data.startTime.toFixed(3));

    // VALIDATION FOR SETTERS
    const handler = {
      set(target, prop, val, proxy) {

        switch (prop) {
          case 'language':
          case 'speaker':
            checkProp(val, 'abbr', prop);
            break;
          case 'endTime': checkTimes(target.startTime, val); break;
          case 'key': checkKey(val); break;
          case 'startTime': checkTimes(val, target.endTime); break;
          case 'url': checkProp(val, 'url', 'url'); break;
          default: break;
        }

        return Reflect.set(target, prop, val, proxy);

      },
    };

    return new Proxy(this, handler);

  }

  toJSON() {
    return Model.simplify(this, [
      'transcription',
      'translation',
      'words',
    ]);
  }

}

module.exports = Phrase;
