const { checkProp } = require('../utilities');
const { Model }     = require('../base');

// don't use destructuring of index.js here - causes problematic circular references
const Morpheme        = require('./Morpheme');
const MultiLangString = require('./MultiLangString');
const Note            = require('./Note');

// checks whether the "key" property is a valid abbreviation
const checkKey = key => {
  if (!/^[(a-z)|(A-Z)|(0-9)]+\.[0-9]{1,3}\.[0-9]{1,2}$/.test(key)) {
    throw new Error(`Improperly formatted key.`);
  }
};

const defaultWord = {
  morphemes:     [],
  transcription: {},
};

/**
 * A class representing a Word token
 * @alias Word
 * @memberof dlx.module:models
 * @prop {Object} gloss         The Leipzig gloss for this word, as a [MultiLangString]{@link dlx.module:models.MultiLangString}.
 * @prop {String} key           The uniquely identifying key for this word token
 * @prop {Array}  morphemes     An array of the morphemes or components in this word token. Each item is a [Morpheme]{@link dlx.module:models.Morpheme}.
 * @prop {Array}  notes         An array of notes about this word token. Each item is a [Note object]{@link dlx.module:models.Note}.
 * @prop {Object} transcription The transcriptions of this word token, optionally in multiple orthographies, as a [MultiLangString]{@link dlx.module:models.MultiLangString}.
 * @prop {String} txn           A shorthand for accessing the default orthography of the `transcription` object, if one is set
 * @prop {Object} translation   The translations of this word token, optionally in multiple orthographies, as a [MultiLangString]{@link dlx.module:models.MultiLangString}. Note that the translation is not the same as a gloss. Use the translation field for free, natural language translations of a word; use the gloss field for Leipzig-style glosses only.
 * @prop {String} tln           A shorthand for accessing the default orthography of the `translation` object, if one is set
 * @prop {String} url           The URL where this word token can be accessed.
 *
 * @example
 * const word = new Word({
 *
 *   transcription: {
 *     spa: 'hablo',
 *     ipa: 'ablo',
 *   },
 *
 *   translation: {
 *     eng: 'I speak',
 *   },
 *
 *   morphemes: [
 *     { lexeme: 'habl' },
 *     { lexeme: 'o' },
 *   ],
 *
 * });
 *
 * console.log(word.transcription.spa); // 'hablo'
 */
class Word extends Model {
  /**
   * Create a new Word
   * @param {Object} [data]             The raw data for this word token
   * @param {Object} [data.gloss]       The gloss for this word, in [MultiLangString format]{@link http://cdn.digitallinguistics.io/schemas/MultiLangString.json}.
   * @param {String} [data.key]         The uniquely identifying key for this word token, as a string. The key for a phrase consists of the abbreviation of the text, a period, the number of phrase within the text, another period, and then the number of this word within the phrase (indexing starts at 1). For example, for the fourth word of the third phrase of a text with the abbreviation `A`, the key would be `A.3.4`. Keys should be unique within a corpus.
   * @param {Array}  data.morphemes     An array of the morphemes or components in this word token. Each morpheme must be formatted as a [Morpheme]{@link http://cdn.digitallinguistics.io/schemas/Morpheme.json}. May be an empty array.
   * @param {Array}  [data.notes]       An array of notes about this word token. Each note must be in the [Note format]{@link http://cdn.digitallinguistics.io/schemas/Note.json}.
   * @param {Object} data.transcription The transcriptions for this word token, optionally in multiple orthographies, formatted as a [MultiLangString]{@link http://cdn.digitallinguistics.io/schemas/MultiLangString.json}.
   * @param {Object} [data.translation] The translations for this word token, optionally in multiple orthographies, formatted as a [MultiLangString]{@link http://cdn.digitallinguistics.io/schemas/MultiLangString.json}. Note that the translation is not the same as a gloss. Use the translation field for free, natural language translations of a word; use the gloss field for Leipzig-style glosses only.
   * @param {String} [data.url]         The URL where this word token can be accessed. Must be a valid URI.
   */
  constructor(data = defaultWord) {

    // VALIDATION
    if (data instanceof Word) return data;

    if (typeof data !== 'object') {
      throw new Error(`The data passed to the Word constructor must be an object if present.`);
    }

    if ('key' in data) checkKey(data.key);

    if ('url' in data) checkProp(data.url, 'url', 'url');

    // INSTANTIATION
    super();

    Object.assign(this, data);

    Model.defineProp(this, 'gloss', MultiLangString, data.gloss || {});
    Model.defineProp(this, 'transcription', MultiLangString, data.transcription || {});
    Model.defineProp(this, 'translation', MultiLangString, data.translation || {});

    Model.defineArray(this, 'morphemes', Morpheme, data.morphemes);
    Model.defineArray(this, 'notes', Note, data.notes);

    MultiLangString.alias(this, 'transcription', 'txn');
    MultiLangString.alias(this, 'translation', 'tln');

    // VALIDATION FOR SETTERS
    return new Proxy(this, {
      set(target, prop, val, proxy) {

        switch (prop) {
          case 'key': checkKey(val); break;
          case 'url': checkProp(val, 'url', 'url'); break;
          default: break;
        }

        return Reflect.set(target, prop, val, proxy);

      },
    });

  }

  toJSON() {
    return Model.simplify(this, [
      'morphemes',
      'transcription',
    ]);
  }

}

module.exports = Word;
