/**
 * A collection of utility functions
 * @module utilities
 * @memberof dlx
 */

const defaultPunctuation = require('./data').punctuation;

/**
 * Creates a new DOM element, and sets its properties
 * @param  {String} type  The type of element to create (e.g. 'div', 'p')
 * @param  {Object} props An object of properties and their values to assign to the new element
 * @return {Object}       Returns the new DOM Node
 */
const createElement = (type, props) => Object.assign(document.createElement(type), props);

/**
 * Returns an array of all the unique characters in a string. Works with both UTF-8 and UTF-16 encoding.
 * @param  {String} string The string to retrieve unique characters from
 * @return {Array}         Returns an array of the unique characters
 * @instance
 */
const getUniqueChars = string => {

  if (typeof string !== 'string') {
    throw new TypeError('The argument passed to "getUniqueChars" must be a string.');
  }

  return Array.from(new Set(string));

};

// filters an array of LexemeReference objects and returns only JSON-unique ones
const getUniqueReferences = (references = []) => {

  if (!Array.isArray(references)) {
    throw new TypeError(`The references must be an Array.`);
  }

  const uniques = [];

  const isEqual = (obj1, obj2) => {

    const stringsEqual = (
      obj1.lexicon === obj2.lexicon
      && obj1.lexeme === obj2.lexeme
      && obj1.relation === obj2.relation
      && obj1.sense === obj2.sense
    );

    if (stringsEqual) {

      if (obj1.variantType && obj2.variantType) {

        return (
          Object.entries(obj1.variantType).every(([key, val]) => val === obj2.variantType[key])
          && Object.entries(obj2.variantType).every(([key, val]) => val === obj1.variantType[key])
        );

      } else if (obj1.variantType || obj2.variantType) {

        return false;

      }

      return true;

    }

    return false;

  };

  references.forEach(ref => {
    if (!uniques.some(obj => isEqual(obj, ref))) uniques.push(ref);
  });

  return uniques;

};

/**
 * Tests whether a string is a valid abbreviation according to the [DLx Abbreviation schema]{@link http://developer.digitallinguistics.io/spec/schemas/abbreviation.html}.
 * @param  {String} string The string to test
 * @return {Boolean}       Returns true if the string is a valid abbreviation, false otherwise
 */
const isAbbr = string => {
  if (typeof string === 'string') return /^[(a-z)|(A-Z)|(0-9)]+$/.test(string);
  return false;
};

/**
 * Tests whether a string is a valid ISO Internet Date-Time string.
 * @param  {String}  string The string to test
 * @return {Boolean}        Returns true if the string is a valid date
 */
const isDateString = string => !isNaN(Date.parse(string));

// check whether an item is iterable
const isIterable = obj => {
  if (!obj) return false;
  return typeof obj[Symbol.iterator] === 'function';
  // NB: cannot use 'in' here because 'obj' might be a string
};

/**
 * Tests whether a string is a valid URI. Code taken from http://stackoverflow.com/questions/8667070/javascript-regular-expression-to-validate-url/15855457#15855457
 * @type {Function}
 * @param {String} string       The string to validate
 * @return {Boolean}            Returns true if the string is a valid URI, false otherwise
 */
const isUrl = string => /^(?:(?:(?:https?|ftp):)?\/\/)(?:\S+(?::\S*)?@)?(?:(?!(?:10|127)(?:\.\d{1,3}){3})(?!(?:169\.254|192\.168)(?:\.\d{1,3}){2})(?!172\.(?:1[6-9]|2\d|3[0-1])(?:\.\d{1,3}){2})(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))|(?:(?:[a-z\u00a1-\uffff0-9]-*)*[a-z\u00a1-\uffff0-9]+)(?:\.(?:[a-z\u00a1-\uffff0-9]-*)*[a-z\u00a1-\uffff0-9]+)*(?:\.(?:[a-z\u00a1-\uffff]{2,})))(?::\d{2,5})?(?:[/?#]\S*)?$/i.test(string);

/**
 * Checks whether an item is valid for the specified type, and returns a generic error message with the provided property name if not.
 * @param  {Any}    item The item to check
 * @param  {String} type The type to check for
 * @param  {String} prop The name of the property to use in the error message
 * @return {Any}         Returns the original value if no error is thrown
 */
const checkProp = (...args) => {

  if (args.length !== 3) throw new Error('Missing arguments to chceckProp function.');

  const [item, type, prop] = args;

  switch (type) {

    case 'abbr': {
      if (!isAbbr(item)) throw new Error(`The "${prop}" property must be a valid abbreviation.`);
      break;
    }

    case 'array': {
      if (!Array.isArray(item)) throw new Error(`The "${prop}" property must be an array.`);
      break;
    }

    case 'date': {
      if (isNaN(Date.parse(item))) throw new Error(`The "${prop}" property must be a valid date.`);
      break;
    }

    case 'integer': {
      if (!Number.isInteger(item)) throw new Error(`The "${prop}" property must be an integer.`);
      break;
    }

    case 'iterable': {
      if (!isIterable(item)) throw new Error(`The "${prop}" property must be an iterable object.`);
      break;
    }

    case 'number': {
      if (typeof item !== 'number') throw new Error(`The "${prop}" property must be a number.`);
      break;
    }

    case 'object': {
      if (typeof item !== 'object') throw new Error(`The "${prop}" property must be an object.`);
      break;
    }

    case 'string': {
      if (typeof item !== 'string') throw new Error(`The "${prop}" property must be a string.`);
      break;
    }

    case 'url': {
      if (!isUrl(item)) throw new Error(`The "${prop}" property must be a valid URI.`);
      break;
    }

    default:

  }

  return item;

};

/**
 * Performs a series of substitutions on a string, replacing unwanted characters with the desired ones (or nothing, meaning the character is simply removed everywhere it occurs).
 * NB: This function cannot properly handle cases where the output of one substitution is also the input of another subsitition (e.g. if you pass it both `ch` => `c` and `c` => `k`).
 * @param  {String} string        The string to sanitize
 * @param  {Object} substitutions An object whose attributes are the characters you wish to replace, and values are the characters you wish to replace them with
 * @return {String}               Returns the sanitized string, with substitutions made
 * @instance
 */
const sanitize = (string, substitutions) => {

  if (typeof string !== 'string') {
    throw new TypeError('The first argument passed to "sanitize" must be a string.');
  }

  if (!(
    typeof substitutions === 'object'
    && Object.values(substitutions).every(val => typeof val === 'string')
    // NB: Object attributes are always strings, so the keys don't need checking
  )) {
    throw new TypeError('The second argument passed to "sanitize" must be an object whose attributes and values are strings.');
  }

  let str = string;
  const badChars = Object.keys(substitutions);

  badChars.sort((a, b) => (a.length < b.length ? 1 : -1));

  badChars.forEach(char => {
    const goodChar = substitutions[char];
    // escape any regular expression characters (regex taken from MDN)
    const badChar  = char.replace(/[.*+?^${}()|[\]\\]/g, '\\$&');
    const regexp   = new RegExp(badChar, 'gu');
    str = str.replace(regexp, goodChar);
  });

  return str;

};

/**
 * Takes an HTML string and converts it to DOM nodes
 * @param  {String} html an HTML string
 * @return {Array}       returns an array of DOM nodes
 */
const toDOM = html => {
  if (html instanceof Node) return [html];
  const template = document.createElement('template');
  template.innerHTML = html;
  return Array.from(template.content.children || []);
};

/**
 * Removes punctuation from a string, and then tokenizes the string, returning an array of tokens.
 * @param  {String} string      The string to tokenize
 * @param  {Array}  delimiters  An array of characters to tokenize the string with
 * @param  {Array}  punctuation An array of punctuation characters to remove from the string
 * @return {Array}              Returns an array of tokens
 * @instance
 */
const tokenize = (string, delimiters, punctuation) => {

  // NB: Don't use default parameters, since they don't allow `null` to be passed.
  const delims = delimiters || ['\\s'];
  const punct  = punctuation || defaultPunctuation;

  if (typeof string !== 'string') {
    throw new TypeError('The "string" argument must be a string.');
  }

  if (!(Array.isArray(delims) && Array.isArray(punct))) {
    throw new TypeError('The "delimiters" and "punctuation" arguments must be arrays.');
  }

  const substitutions = {};

  punct.forEach(item => { substitutions[item] = ''; });

  const depunctuated = sanitize(string, substitutions);
  const pattern      = `[${delims.join('')}]+`;
  const regexp       = new RegExp(pattern, 'gu');

  return depunctuated.split(regexp);

};

/**
 * Converts a string from one writing system to another.
 * @param {String} string         The string to transliterate.
 * @param {Object} substitutions  A hash of substitutions to make on the string. Each key in the object should be a string to replace, and its value should be the string to replace it with.
 * @return {String}               Returns the transliterated string
 * @instance
 */
const transliterate = (string, substitutions) => {

  if (typeof string !== 'string') {
    throw new TypeError('The first argument passed to "transliterate" must be a string.');
  }

  if (!(
    typeof substitutions === 'object'
    && Object.values(substitutions).every(val => typeof val === 'string')
  )) {
    throw new TypeError('The second argument passed to "substitutions" must be an object whose attributes and values are strings.');
  }

  const duplicates = new Map;
  const subs       = new Map;
  let str          = string;

  const inputs = Object.keys(substitutions).sort((a, b) => (a.length < b.length ? 1 : -1));
  inputs.forEach(input => subs.set(input, substitutions[input]));

  subs.forEach((replacement, input) => {

    // check whether any substitution overlaps with another substitution
    // temporarily replace the duplicate substitution with a random number if so
    if (inputs.includes(replacement)) {
      const temp = Math.floor(Math.random() * 100000);
      duplicates.set(temp, replacement);
      subs.set(input, temp);
    }

    // make the substitutions using the temporary placeholders
    const regexp = new RegExp(input, 'gu');
    str = str.replace(regexp, subs.get(input));

  });

  // now replace the temporary placeholders with their proper values
  duplicates.forEach((replacement, temp) => {
    const regexp = new RegExp(temp, 'gu');
    str = str.replace(regexp, replacement);
  });

  return str;

};

module.exports = {
  checkProp,
  createElement,
  getUniqueChars,
  getUniqueReferences,
  isAbbr,
  isDateString,
  isIterable,
  isUrl,
  sanitize,
  toDOM,
  tokenize,
  transliterate,
};
