/* global dlx */

/* eslint-disable
  func-names,
  max-nested-callbacks,
  max-statements,
  no-magic-numbers,
  no-shadow,
  object-curly-newline,
  object-property-newline,
  prefer-arrow-callback
*/

describe('dlx.models.Phrase', function() {

  const { Model } = dlx.base;

  const {
    MultiLangString,
    Note,
    Phrase,
    Tags,
    Word,
  } = dlx.models;

  const data    = {
    transcription: {
      ipa: 'ola me jamo dænjɛl',
      spa: 'Hola, me llamo Daniel.',
    },
    translation:   {
      eng:  'Hello, my name is Daniel.',
      type: 'free',
    },
    words:         [
      {
        morphemes:     [],
        transcription: {},
        translation:   {},
      },
      {
        morphemes:     [],
        transcription: {},
        translation:   {},
      },
      {
        morphemes:     [],
        transcription: {},
        translation:   {},
      },
      {
        morphemes:     [],
        transcription: {},
        translation:   {},
      },
    ],
  };

  const phrase = new Phrase(data);

  it('Errors', function() {

    const badData     = () => new Phrase(true);
    const badTln      = () => new Phrase(Object.assign({}, data, { translation: true }));
    const badTxn      = () => new Phrase(Object.assign({}, data, { transcription: true }));
    const badWords    = () => new Phrase(Object.assign({}, data, { words: true }));
    const noEndTime   = () => new Phrase(Object.assign({}, data, { startTime: 10.00 }));
    const noStartTime = () => new Phrase(Object.assign({}, data, { endTime: 20.00 }));
    const timeOverlap = () => new Phrase(Object.assign({}, data, { endTime: 10.00, startTime: 20.00 }));

    expect(badData).toThrow();
    expect(badTln).toThrow();
    expect(badTxn).toThrow();
    expect(badWords).toThrow();
    expect(noEndTime).toThrow();
    expect(noStartTime).toThrow();
    expect(timeOverlap).toThrow();

  });

  it('new Phrase()', function() {

    expect(phrase instanceof Model).toBe(true);

    const noData = () => new Phrase();
    expect(noData).not.toThrow();

    const oldPhrase = new Phrase;
    const newPhrase = new Phrase(oldPhrase);
    expect(newPhrase).toBe(oldPhrase);

  });

  it('Phrase.prototype.{customProperty}', function() {
    const phrase = new Phrase(Object.assign({}, data, { customProperty1: 'custom value 1' }));
    phrase.customProperty2 = 'custom value 2';
    expect(phrase.customProperty1).toBe('custom value 1');
    expect(phrase.customProperty2).toBe('custom value 2');
  });

  it('Phrase.prototype.endTime', function() {
    const phrase = new Phrase(Object.assign({}, data, {
      endTime:   20.000,
      startTime: 10.000,
    }));
    const badEndTime = () => new Phrase(Object.assign({}, data, {
      endTime:   true,
      startTime: 10.000,
    }));
    const correctFormatPhrase = new Phrase(Object.assign({}, data, {
      endTime:   20.000,
      startTime: 10.000,
    }));
    const incorrectFormatPhrase = new Phrase(Object.assign({}, data, {
      endTime:   20.0012345,
      startTime: 10.000,
    }));
    const setBadEndTime = () => { phrase.endTime = 5.000; };
    expect(badEndTime).toThrow();
    expect(correctFormatPhrase.endTime).toBe(20.000);
    expect(incorrectFormatPhrase.endTime).toBe(20.001);
    expect(setBadEndTime).toThrow();
  });

  it('Phrase.prototype.key', function() {
    const key       = 'A.3';
    const phrase    = new Phrase(Object.assign({}, data, { key }));
    const badKey    = () => new Phrase(Object.assign({}, data, { key: '11' }));
    const setBadKey = () => { phrase.key = '11'; };
    expect(phrase.key).toBe('A.3');
    expect(badKey).toThrow();
    expect(setBadKey).toThrow();
  });

  it('Phrase.prototype.language', function() {
    const language   = 'ctm';
    const phrase     = new Phrase(Object.assign({}, data, { language }));
    const badLang    = () => new Phrase(Object.assign({}, data, { language: 'Tlahuapa Mixtec' }));
    const setBadLang = () => { phrase.language = 'Tlahuapa Mixtec'; };
    expect(phrase.language).toBe(language);
    expect(badLang).toThrow();
    expect(setBadLang).toThrow();
  });

  it('Phrase.prototype.notes', function() {
    const badNotes = () => new Phrase(Object.assign({}, data, { notes: true }));
    const note = new Note('test');
    const phrase = new Phrase(Object.assign({}, data, { notes: [note] }));
    expect(badNotes).toThrow();
    expect(phrase.notes[0]).toBe(note);
  });

  it('Phrase.prototype.speaker', function() {
    const badSpeaker = () => new Phrase(Object.assign({}, data, { speaker: 'Daniel Hieber' }));
    const setBadSpeaker = () => { phrase.speaker = 'Daniel Hieber'; };
    expect(badSpeaker).toThrow();
    expect(setBadSpeaker).toThrow();
  });

  it('Phrase.prototype.startTime', function() {
    const phrase = new Phrase(Object.assign({}, data, {
      endTime:   20.000,
      startTime: 10.000,
    }));
    const badStartTime = () => new Phrase(Object.assign({}, data, {
      endTime:   20.000,
      startTime: true,
    }));
    const correctFormatPhrase = new Phrase(Object.assign({}, data, {
      endTime:   20.000,
      startTime: 10.000,
    }));
    const incorrectFormatPhrase = new Phrase(Object.assign({}, data, {
      endTime:   20.000,
      startTime: 10.0012345,
    }));
    const setBadStartTime = () => { phrase.startTime = 25.000; };
    expect(badStartTime).toThrow();
    expect(correctFormatPhrase.startTime).toBe(10.000);
    expect(incorrectFormatPhrase.startTime).toBe(10.001);
    expect(setBadStartTime).toThrow();
  });

  it('Phrase.prototype.tags', function() {
    const badTags = () => new Phrase(Object.assign({}, data, { tags: true }));
    phrase.tags = {};
    expect(phrase.tags instanceof Tags).toBe(true);
    expect(badTags).toThrow();
  });

  it('Phrase.prototype.toJSON', function() {
    const phrase = new Phrase(Object.assign({ tags: {} }, data));
    phrase.tags.set('test', true);
    const pojo = JSON.parse(JSON.stringify(phrase));
    expect(pojo.tags.test).toBe(true);
  });

  it('Phrase.prototype.transcription', function() {
    expect(phrase.transcription instanceof MultiLangString).toBe(true);
    expect(phrase.transcription.spa).toBe(data.transcription.spa);
  });

  it('Phrase.prototype.translation', function() {
    expect(phrase.translation instanceof MultiLangString).toBe(true);
    expect(phrase.translation.eng).toBe(data.translation.eng);
    expect(phrase.translation.type).toBe('free');
  });

  it('Phrase.prototype.url', function() {
    const url = 'https://digitallinguistics.io';
    const phrase = new Phrase(Object.assign({}, data, { url }));
    const badUrl = () => new Phrase(Object.assign({}, data, { url: 'digitallinguistics.io' }));
    expect(badUrl).toThrow();
    expect(phrase.url).toBe(url);
  });

  it('Phrase.prototype.words', function() {
    expect(Array.isArray(phrase.words)).toBe(true);
    expect(phrase.words.length).toBe(data.words.length);
    expect(phrase.words.every(word => word instanceof Word)).toBe(true);
  });

});
