/* global dlx */

/* eslint-disable
  func-names,
  max-nested-callbacks,
  no-magic-numbers,
  object-curly-newline,
  object-property-newline,
  prefer-arrow-callback
*/

describe('dlx.utils', function() {

  const utils = dlx.utils;

  describe('getUniqueChars', function() {

    const getUniqueChars = utils.getUniqueChars;

    it('dlx.utils.getUniqueChars === "function"', function() {
      expect(typeof getUniqueChars).toBe('function');
    });

    it('Error: The argument passed to "getUniqueChars" must be a string.', function() {

      const message = 'The argument passed to "getUniqueChars" must be a string.';
      const passNothing = () => getUniqueChars();
      const passNum = () => getUniqueChars(42);
      const passEmptyStr = () => getUniqueChars('');

      expect(passNothing).toThrowError(message);
      expect(passNum).toThrowError(message);
      expect(passEmptyStr).not.toThrow();

    });

    it('"ñ" → ["ñ"] (non-combining)', function() {
      const enye = 'ñ';
      expect(getUniqueChars(enye)[0]).toBe(enye);
    });

    it('"ñ" → ["n", "̃"] (combining)', function() {
      const combiningEnye = 'ñ';
      expect(getUniqueChars(combiningEnye).length).toBe(2);
      expect(getUniqueChars(combiningEnye)[1]).toBe('̃');
    });

    it('"𐐷" → ["𐐷"] (UTF-16)', function() {
      const yee = '𐐷';
      expect(getUniqueChars(yee)[0]).toBe(yee);
    });

  });

  describe('isAbbr', function() {

    const isAbbr = utils.isAbbr;

    it('returns false for invalid abbreviations', function() {
      expect(isAbbr('eng.2')).toBe(false);
      expect(isAbbr('eng american')).toBe(false);
    });

    it('returns true for valid abbreviations', function() {
      expect(isAbbr('eng')).toBe(true);
    });

  });

  describe('isDateString', function() {

    const isDateString = utils.isDateString;

    it('returns false for invalid dates', function() {
      expect(isDateString('today')).toBe(false);
    });

    it('returns truee for valid dates', function() {
      expect(isDateString('2017-02-10')).toBe(true);
    });

  });

  describe('isUrl', function() {

    const isUrl = utils.isUrl;

    it('returns false for invalid URLs', function() {
      expect(isUrl('digitallinguistics.io')).toBe(false);
    });

    it('returns true for valid URLs', function() {
      expect(isUrl('http://digitallinguistics.io/')).toBe(true);
    });

  });

  describe('sanitize', function() {

    const sanitize = utils.sanitize;

    const subs = {
      b: 'v',         // plain old substitution
      ñ: 'ñ',         // combining → non-combining
      ó: 'ò',         // change diacritics
      𐐷: 'y',        // UTF-16
    };

    const badStr = 'baca bbaca eñye eññye óh óóh 𐐷ee 𐐷𐐷ee';
    const goodStr = 'vaca vvaca eñye eññye òh òòh yee yyee';

    it('dlx.utils.sanitize === "function"', function() {
      expect(typeof sanitize).toBe('function');
    });

    it('Error: The first argument passed to "sanitize" must be a string.', function() {

      const passNothing = () => sanitize();
      const passNum = () => sanitize(42, subs);
      const passObj = () => sanitize({}, subs);
      const passEmptyStr = () => sanitize('', subs);
      const message = 'The first argument passed to "sanitize" must be a string.';

      expect(passNothing).toThrowError(message);
      expect(passNum).toThrowError(message);
      expect(passObj).toThrowError(message);
      expect(passEmptyStr).not.toThrow();

    });

    it('Error: The second argument passed to "sanitize" must be an object whose attribute values are strings.', function() {

      // NB: Object attributes are always strings, so the keys don't need checking
      const passStr = () => sanitize(badStr, badStr);
      const passBadVals = () => sanitize(badStr, { a: 42, b: '42' });
      const passArr = () => sanitize(badStr, ['b', 'v']);
      const passSymbol = () => sanitize(badStr, { [Symbol('text')]: 'text', b: 'text' });
      const message = 'The second argument passed to "sanitize" must be an object whose attributes and values are strings.';

      expect(passStr).toThrowError(message);
      expect(passBadVals).toThrowError(message);
      expect(passArr).not.toThrow();
      expect(passSymbol).not.toThrow();

    });

    it('makes correct substitutions', function() {
      expect(sanitize(badStr, subs)).toBe(goodStr);
    });

  });

  describe('tokenize', function DLxSpec() {

    const string   = 'Hello! Are you my mother?';
    const tokenize = utils.tokenize;

    it('dlx.utils.tokenize === "function"', function exists() {
      expect(tokenize).toBeDefined();
    });

    it('Error: The "string" argument must be a string.', function firstArg() {

      const withString = () => tokenize('hello', [], []);
      const withObject = () => tokenize({}, [], []);
      const withNothing = () => tokenize(null, [], []);

      expect(withString).not.toThrow();
      expect(withObject).toThrow();
      expect(withNothing).toThrow();

    });

    it('Error: The "delimiters" and "punctuation" arguments must be arrays. (1st argument)', function secondArg() {

      const withArray = () => tokenize('hello', [], []);
      const withObject = () => tokenize('hello', {}, []);
      const withNothing = () => tokenize('hello', null, []);

      expect(withArray).not.toThrow();
      expect(withObject).toThrow();
      expect(withNothing).not.toThrow();

    });

    it('Error: The "delimiters" and "punctuation" arguments must be arrays. (2nd argument)', function thirdArg() {

      const withObject = () => tokenize('hello', [], {});
      const withNothing = () => tokenize('hello', [], null);

      expect(withObject).toThrow();
      expect(withNothing).not.toThrow();

    });

    it('uses default delimiters and punctuation', function defaults() {

      const stringOnly = () => tokenize('hello');
      const tokens = tokenize(string);

      expect(stringOnly).not.toThrow();
      expect(tokens.length).toBe(5);
      expect(tokens[0]).toBe('Hello');
      expect(tokens[4]).toBe('mother');

    });

    it('removes punctuation', function() {

      const depunctuated = tokenize(string, [], ['!', '?']);
      const hasPunctuation = depunctuated.some(str => str.includes('!') || str.includes('?'));

      expect(Array.isArray(depunctuated)).toBe(true);
      expect(depunctuated.length).toBe(1);
      expect(depunctuated[0]).toBe('Hello Are you my mother');
      expect(hasPunctuation).toBe(false);

    });

    it('tokenizes', function() {
      expect(tokenize(string, [' '], ['!', '?']).length).toBe(5);
    });

  });

  describe('transliterate', function() {

    const transliterate = utils.transliterate;
    const string        = 'čaːd ki ǯaʔa ʔap čuyi. him caspik.';
    const correctString = 'caad ki jaqa qap cuyi. him tsaspik.'
    const substitutions = {
      aː: 'aa',
      c:  'ts',
      č:  'c',
      ǯ:  'j',
      ʔ:  'q',
    };

    it('Errors', function() {
      expect(transliterate).toThrow();
      expect(() => transliterate(string)).toThrow();
      expect(() => transliterate(string, {})).not.toThrow();
      expect(() => transliterate(string, { c: true })).toThrow();
    });

    it('transliterates', function() {
      expect(transliterate(string, substitutions)).toBe(correctString);
    });

  });

});
